/** 
 * @file llmd5.cpp
 *
 * $LicenseInfo:firstyear=2001&license=viewerlgpl$
 * Second Life Viewer Source Code
 * Copyright (C) 2010, Linden Research, Inc.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * version 2.1 of the License only.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 * Linden Research, Inc., 945 Battery Street, San Francisco, CA  94111  USA
 * $/LicenseInfo$
 */

// llMD5.CC - source code for the C++/object oriented translation and 
//          modification of MD5.
//
// Adapted to Linden Lab by Frank Filipanits, 6/25/2002
// Fixed potential memory leak, James Cook, 6/27/2002

// Translation and modification (c) 1995 by Mordechai T. Abzug 

// This translation/ modification is provided "as is," without express or 
// implied warranty of any kind.

// The translator/ modifier does not claim (1) that MD5 will do what you think 
// it does; (2) that this translation/ modification is accurate; or (3) that 
// this software is "merchantible."  (Language for this disclaimer partially 
// copied from the disclaimer below).

/* based on:

   MD5C.C - RSA Data Security, Inc., MD5 message-digest algorithm
   MDDRIVER.C - test driver for MD2, MD4 and MD5


   Copyright (C) 1991-2, RSA Data Security, Inc. Created 1991. All
rights reserved.

License to copy and use this software is granted provided that it
is identified as the "RSA Data Security, Inc. MD5 Message-Digest
Algorithm" in all material mentioning or referencing this software
or this function.

License is also granted to make and use derivative works provided
that such works are identified as "derived from the RSA Data
Security, Inc. MD5 Message-Digest Algorithm" in all material
mentioning or referencing the derived work.

RSA Data Security, Inc. makes no representations concerning either
the merchantability of this software or the suitability of this
software for any particular purpose. It is provided "as is"
without express or implied warranty of any kind.

These notices must be retained in any copies of any part of this
documentation and/or software.

 */





#include "linden_common.h"

#include "llmd5.h"

#include <iostream>		// cerr

// how many bytes to grab at a time when checking files
const int LLMD5::BLOCK_LEN = 4096;


// LLMD5 simple initialization method

LLMD5::LLMD5()
{
  init();
}




// MD5 block update operation. Continues an MD5 message-digest
// operation, processing another message block, and updating the
// context.

void LLMD5::update (const uint1 *input, const uint4 input_length) {

  uint4 input_index, buffer_index;
  uint4 buffer_space;                // how much space is left in buffer

  if (mFinalized){  // so we can't update!
	  std::cerr << "LLMD5::update:  Can't update a finalized digest!" << std::endl;
    return;
  }

  // Compute number of bytes mod 64
  buffer_index = (unsigned int)((mCount[0] >> 3) & 0x3F);

  // Update number of bits
  if (  (mCount[0] += ((uint4) input_length << 3))<((uint4) input_length << 3) )
	  mCount[1]++;

  mCount[1] += ((uint4)input_length >> 29);


  buffer_space = 64 - buffer_index;  // how much space is left in buffer

  // now, transform each 64-byte piece of the input, bypassing the buffer
  if (input == NULL || input_length == 0){
	  std::cerr << "LLMD5::update:  Invalid input!" << std::endl;
	  return;
  }

  // Transform as many times as possible.
  if (input_length >= buffer_space) { // ie. we have enough to fill the buffer
    // fill the rest of the buffer and transform
    memcpy(	/* Flawfinder: ignore */
		mBuffer + buffer_index,
		input,
		buffer_space);
    transform (mBuffer);

    for (input_index = buffer_space; input_index + 63 < input_length; 
	 input_index += 64)
      transform (input+input_index);

    buffer_index = 0;  // so we can buffer remaining
  }
  else
    input_index=0;     // so we can buffer the whole input


  // and here we do the buffering:
  memcpy(mBuffer+buffer_index, input+input_index, input_length-input_index);		/* Flawfinder: ignore */
}



// MD5 update for files.
// Like above, except that it works on files (and uses above as a primitive.)

void LLMD5::update(FILE* file){

  unsigned char buffer[BLOCK_LEN];		/* Flawfinder: ignore */
  int len;

  while ( (len=(int)fread(buffer, 1, BLOCK_LEN, file)) )
    update(buffer, len);

  fclose (file);

}

// MD5 update for istreams.
// Like update for files; see above.

void LLMD5::update(std::istream& stream){

  unsigned char buffer[BLOCK_LEN];		/* Flawfinder: ignore */
  int len;

  while (stream.good()){
    stream.read( (char*)buffer, BLOCK_LEN); 	/* Flawfinder: ignore */		// note that return value of read is unusable.
    len=(int)stream.gcount();
    update(buffer, len);
  }

}

void  LLMD5::update(const std::string& s)
{
	update((unsigned char *)s.c_str(),s.length());
}

// MD5 finalization. Ends an MD5 message-digest operation, writing the
// the message digest and zeroizing the context.


void LLMD5::finalize (){

  unsigned char bits[8];		/* Flawfinder: ignore */
  unsigned int index, padLen;
  static uint1 PADDING[64]={
    0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    };

  if (mFinalized){
    std::cerr << "LLMD5::finalize:  Already finalized this digest!" << std::endl;
    return;
  }

  // Save number of bits
  encode (bits, mCount, 8);

  // Pad out to 56 mod 64.
  index = (uint4) ((mCount[0] >> 3) & 0x3f);
  padLen = (index < 56) ? (56 - index) : (120 - index);
  update (PADDING, padLen);

  // Append length (before padding)
  update (bits, 8);

  // Store state in digest
  encode (mDigest, mState, 16);

  // Zeroize sensitive information
  memset (mBuffer, 0, sizeof(*mBuffer));

  mFinalized=true;

}




LLMD5::LLMD5(FILE *file){

  init();  // must be called be all constructors
  update(file);
  finalize ();
}




LLMD5::LLMD5(std::istream& stream){

  init();  // must called by all constructors
  update (stream);
  finalize();
}

// Digest a string of the format ("%s:%i" % (s, number))
LLMD5::LLMD5(const unsigned char *string, const unsigned int number)
{
	const char *colon = ":";
	char tbuf[16];		/* Flawfinder: ignore */
	init();
	update(string, (U32)strlen((const char *) string));		/* Flawfinder: ignore */
	update((const unsigned char *) colon, (U32)strlen(colon));		/* Flawfinder: ignore */
	snprintf(tbuf, sizeof(tbuf), "%i", number);	/* Flawfinder: ignore */
	update((const unsigned char *) tbuf, (U32)strlen(tbuf));	/* Flawfinder: ignore */
	finalize();
}

// Digest a string
LLMD5::LLMD5(const unsigned char *s)
{
	init();
	update(s, (U32)strlen((const char *) s));		/* Flawfinder: ignore */
	finalize();
}

void LLMD5::raw_digest(unsigned char *s) const
{
	if (!mFinalized)
	{
		std::cerr << "LLMD5::raw_digest:  Can't get digest if you haven't "<<
			"finalized the digest!" << std::endl;
		s[0] = '\0';
		return;
	}

	memcpy(s, mDigest, 16);		/* Flawfinder: ignore */
	return;
}



void LLMD5::hex_digest(char *s) const
{
	int i;

	if (!mFinalized)
	{
		std::cerr << "LLMD5::hex_digest:  Can't get digest if you haven't "<<
		  "finalized the digest!" <<std::endl;
		s[0] = '\0';
		return;
	}

	for (i=0; i<16; i++)
	{
		sprintf(s+i*2, "%02x", mDigest[i]);		/* Flawfinder: ignore */
	}

	s[32]='\0';

	return;
}






std::ostream& operator<<(std::ostream &stream, LLMD5 context)
{
	char s[33];		/* Flawfinder: ignore */
	context.hex_digest(s);
	stream << s;
	return stream;
}

bool operator==(const LLMD5& a, const LLMD5& b)
{
	unsigned char a_guts[16];
	unsigned char b_guts[16];
	a.raw_digest(a_guts);
	b.raw_digest(b_guts);
	if (memcmp(a_guts,b_guts,16)==0)
		return true;
	else
		return false;
}

bool operator!=(const LLMD5& a, const LLMD5& b)
{
	return !(a==b);
}

// PRIVATE METHODS:

void LLMD5::init(){
  mFinalized=false;  // we just started!

  // Nothing counted, so count=0
  mCount[0] = 0;
  mCount[1] = 0;

  // Load magic initialization constants.
  mState[0] = 0x67452301;
  mState[1] = 0xefcdab89;
  mState[2] = 0x98badcfe;
  mState[3] = 0x10325476;
}



// Constants for MD5Transform routine.
// Although we could use C++ style constants, defines are actually better,
// since they let us easily evade scope clashes.

#define S11 7
#define S12 12
#define S13 17
#define S14 22
#define S21 5
#define S22 9
#define S23 14
#define S24 20
#define S31 4
#define S32 11
#define S33 16
#define S34 23
#define S41 6
#define S42 10
#define S43 15
#define S44 21

// #defines are faster then inline, etc because the compiler is not required to inline.
// Timing tests prove that this works ~40% faster on win with msvc++2k3 over using static inline.

/* F, G, H and I are basic MD5 functions.
 */
#define F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define G(x, y, z) (((x) & (z)) | ((y) & (~z)))
#define H(x, y, z) ((x) ^ (y) ^ (z))
#define I(x, y, z) ((y) ^ ((x) | (~z)))

/* ROTATE_LEFT rotates x left n bits.
 */
#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))

/* FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
Rotation is separate from addition to prevent recomputation.
 */
#define FF(a, b, c, d, x, s, ac) { \
 (a) += F ((b), (c), (d)) + (x) + (U32)(ac); \
 (a) = ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define GG(a, b, c, d, x, s, ac) { \
 (a) += G ((b), (c), (d)) + (x) + (U32)(ac); \
 (a) = ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define HH(a, b, c, d, x, s, ac) { \
 (a) += H ((b), (c), (d)) + (x) + (U32)(ac); \
 (a) = ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define II(a, b, c, d, x, s, ac) { \
 (a) += I ((b), (c), (d)) + (x) + (U32)(ac); \
 (a) = ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }



// LLMD5 basic transformation. Transforms state based on block.
void LLMD5::transform (const uint1 *buffer)
{

  uint4 a = mState[0], b = mState[1], c = mState[2], d = mState[3], x[16];

  decode (x, buffer, 64);

  assert(!mFinalized);  // not just a user error, since the method is private

  /* Round 1 */
  FF (a, b, c, d, x[ 0], S11, 0xd76aa478); /* 1 */
  FF (d, a, b, c, x[ 1], S12, 0xe8c7b756); /* 2 */
  FF (c, d, a, b, x[ 2], S13, 0x242070db); /* 3 */
  FF (b, c, d, a, x[ 3], S14, 0xc1bdceee); /* 4 */
  FF (a, b, c, d, x[ 4], S11, 0xf57c0faf); /* 5 */
  FF (d, a, b, c, x[ 5], S12, 0x4787c62a); /* 6 */
  FF (c, d, a, b, x[ 6], S13, 0xa8304613); /* 7 */
  FF (b, c, d, a, x[ 7], S14, 0xfd469501); /* 8 */
  FF (a, b, c, d, x[ 8], S11, 0x698098d8); /* 9 */
  FF (d, a, b, c, x[ 9], S12, 0x8b44f7af); /* 10 */
  FF (c, d, a, b, x[10], S13, 0xffff5bb1); /* 11 */
  FF (b, c, d, a, x[11], S14, 0x895cd7be); /* 12 */
  FF (a, b, c, d, x[12], S11, 0x6b901122); /* 13 */
  FF (d, a, b, c, x[13], S12, 0xfd987193); /* 14 */
  FF (c, d, a, b, x[14], S13, 0xa679438e); /* 15 */
  FF (b, c, d, a, x[15], S14, 0x49b40821); /* 16 */

 /* Round 2 */
  GG (a, b, c, d, x[ 1], S21, 0xf61e2562); /* 17 */
  GG (d, a, b, c, x[ 6], S22, 0xc040b340); /* 18 */
  GG (c, d, a, b, x[11], S23, 0x265e5a51); /* 19 */
  GG (b, c, d, a, x[ 0], S24, 0xe9b6c7aa); /* 20 */
  GG (a, b, c, d, x[ 5], S21, 0xd62f105d); /* 21 */
  GG (d, a, b, c, x[10], S22,  0x2441453); /* 22 */
  GG (c, d, a, b, x[15], S23, 0xd8a1e681); /* 23 */
  GG (b, c, d, a, x[ 4], S24, 0xe7d3fbc8); /* 24 */
  GG (a, b, c, d, x[ 9], S21, 0x21e1cde6); /* 25 */
  GG (d, a, b, c, x[14], S22, 0xc33707d6); /* 26 */
  GG (c, d, a, b, x[ 3], S23, 0xf4d50d87); /* 27 */
  GG (b, c, d, a, x[ 8], S24, 0x455a14ed); /* 28 */
  GG (a, b, c, d, x[13], S21, 0xa9e3e905); /* 29 */
  GG (d, a, b, c, x[ 2], S22, 0xfcefa3f8); /* 30 */
  GG (c, d, a, b, x[ 7], S23, 0x676f02d9); /* 31 */
  GG (b, c, d, a, x[12], S24, 0x8d2a4c8a); /* 32 */

  /* Round 3 */
  HH (a, b, c, d, x[ 5], S31, 0xfffa3942); /* 33 */
  HH (d, a, b, c, x[ 8], S32, 0x8771f681); /* 34 */
  HH (c, d, a, b, x[11], S33, 0x6d9d6122); /* 35 */
  HH (b, c, d, a, x[14], S34, 0xfde5380c); /* 36 */
  HH (a, b, c, d, x[ 1], S31, 0xa4beea44); /* 37 */
  HH (d, a, b, c, x[ 4], S32, 0x4bdecfa9); /* 38 */
  HH (c, d, a, b, x[ 7], S33, 0xf6bb4b60); /* 39 */
  HH (b, c, d, a, x[10], S34, 0xbebfbc70); /* 40 */
  HH (a, b, c, d, x[13], S31, 0x289b7ec6); /* 41 */
  HH (d, a, b, c, x[ 0], S32, 0xeaa127fa); /* 42 */
  HH (c, d, a, b, x[ 3], S33, 0xd4ef3085); /* 43 */
  HH (b, c, d, a, x[ 6], S34,  0x4881d05); /* 44 */
  HH (a, b, c, d, x[ 9], S31, 0xd9d4d039); /* 45 */
  HH (d, a, b, c, x[12], S32, 0xe6db99e5); /* 46 */
  HH (c, d, a, b, x[15], S33, 0x1fa27cf8); /* 47 */
  HH (b, c, d, a, x[ 2], S34, 0xc4ac5665); /* 48 */

  /* Round 4 */
  II (a, b, c, d, x[ 0], S41, 0xf4292244); /* 49 */
  II (d, a, b, c, x[ 7], S42, 0x432aff97); /* 50 */
  II (c, d, a, b, x[14], S43, 0xab9423a7); /* 51 */
  II (b, c, d, a, x[ 5], S44, 0xfc93a039); /* 52 */
  II (a, b, c, d, x[12], S41, 0x655b59c3); /* 53 */
  II (d, a, b, c, x[ 3], S42, 0x8f0ccc92); /* 54 */
  II (c, d, a, b, x[10], S43, 0xffeff47d); /* 55 */
  II (b, c, d, a, x[ 1], S44, 0x85845dd1); /* 56 */
  II (a, b, c, d, x[ 8], S41, 0x6fa87e4f); /* 57 */
  II (d, a, b, c, x[15], S42, 0xfe2ce6e0); /* 58 */
  II (c, d, a, b, x[ 6], S43, 0xa3014314); /* 59 */
  II (b, c, d, a, x[13], S44, 0x4e0811a1); /* 60 */
  II (a, b, c, d, x[ 4], S41, 0xf7537e82); /* 61 */
  II (d, a, b, c, x[11], S42, 0xbd3af235); /* 62 */
  II (c, d, a, b, x[ 2], S43, 0x2ad7d2bb); /* 63 */
  II (b, c, d, a, x[ 9], S44, 0xeb86d391); /* 64 */

  mState[0] += a;
  mState[1] += b;
  mState[2] += c;
  mState[3] += d;

  // Zeroize sensitive information.
  memset ( (uint1 *) x, 0, sizeof(x));

}



// Encodes input (UINT4) into output (unsigned char). Assumes len is
// a multiple of 4.
void LLMD5::encode (uint1 *output, const uint4 *input, const uint4 len) {

  unsigned int i, j;

  for (i = 0, j = 0; j < len; i++, j += 4) {
    output[j]   = (uint1)  (input[i] & 0xff);
    output[j+1] = (uint1) ((input[i] >> 8) & 0xff);
    output[j+2] = (uint1) ((input[i] >> 16) & 0xff);
    output[j+3] = (uint1) ((input[i] >> 24) & 0xff);
  }
}




// Decodes input (unsigned char) into output (UINT4). Assumes len is
// a multiple of 4.
void LLMD5::decode (uint4 *output, const uint1 *input, const uint4 len){

  unsigned int i, j;

  for (i = 0, j = 0; j < len; i++, j += 4)
    output[i] = ((uint4)input[j]) | (((uint4)input[j+1]) << 8) |
      (((uint4)input[j+2]) << 16) | (((uint4)input[j+3]) << 24);
}


